@@ -1,4 +1,5 @@
 from ._channel import Channel
+from ._grid import AnyGridPlan, GridFromEdges, GridRelative, NoGrid
 from ._mda_event import MDAEvent, PropertyTuple
 from ._mda_sequence import MDASequence
 from ._position import Position
@@ -19,6 +20,10 @@ from ._z import (
 )
 
 __all__ = [
+    "AnyGridPlan",
+    "GridFromEdges",
+    "GridRelative",
+    "NoGrid",
     "Channel",
     "MDAEvent",
     "MDASequence",
