@@ -2,7 +2,6 @@ from __future__ import annotations
 
 from itertools import product
 from typing import (
-    TYPE_CHECKING,
     Any,
     Dict,
     Iterator,
@@ -10,6 +9,7 @@ from typing import (
     Sequence,
     Tuple,
     Union,
+    cast,
     no_type_check,
 )
 from uuid import UUID, uuid4
@@ -18,22 +18,21 @@ from warnings import warn
 import numpy as np
 from pydantic import Field, PrivateAttr, root_validator, validator
 
+from . import _mda_event
 from ._base_model import UseqModel
 from ._channel import Channel
+from ._grid import AnyGridPlan, GridPosition, NoGrid
 from ._mda_event import MDAEvent
 from ._position import Position
 from ._time import AnyTimePlan, NoT
 from ._z import AnyZPlan, NoZ
 
-if TYPE_CHECKING:
-    from ._time import NoT
-    from ._z import NoZ
-
 TIME = "t"
 CHANNEL = "c"
 POSITION = "p"
 Z = "z"
-INDICES = (TIME, POSITION, CHANNEL, Z)
+GRID = "g"
+INDICES = (TIME, POSITION, GRID, CHANNEL, Z)
 
 Undefined = object()
 
@@ -52,11 +51,13 @@ class MDASequence(UseqModel):
     metadata : dict
         A dictionary of user metadata to be stored with the sequence.
     axis_order : str
-        The order of the axes in the sequence. Must be a permutation of `"tpcz"`. The
-        default is `"tpcz"`.
+        The order of the axes in the sequence. Must be a permutation of `"tpgcz"`. The
+        default is `"tpgcz"`.
     stage_positions : tuple[Position, ...]
-        The stage positions to visit. (each with `x`, `y`, `z`, `name`, and `z_plan`,
+        The stage positions to visit. (each with `x`, `y`, `z`, `name`, and `sequence`,
         all of which are optional).
+    grid_plan : GridFromEdges, GridRelative, NoGrid
+        The grid plan to follow. One of `GridFromEdges`, `GridRelative` or `NoGrid`.
     channels : tuple[Channel, ...]
         The channels to acquire. see `Channel`.
     time_plan : MultiPhaseTimePlan | TIntervalDuration | TIntervalLoops \
@@ -77,11 +78,12 @@ class MDASequence(UseqModel):
     >>> seq = MDASequence(
     ...     time_plan={"interval": 0.1, "loops": 2},
     ...     stage_positions=[(1, 1, 1)],
+    ...     grid_plan={"rows": 2, "cols": 2},
     ...     z_plan={"range": 3, "step": 1},
     ...     channels=[{"config": "DAPI", "exposure": 1}]
     ... )
     >>> print(seq)
-    Multi-Dimensional Acquisition ▶ nt: 2, np: 1, nc: 1, nz: 4
+    Multi-Dimensional Acquisition ▶ nt: 2, np: 1, nc: 1, nz: 4, ng: 4
 
     >>> for event in seq:
     ...     print(event)
@@ -90,6 +92,9 @@ class MDASequence(UseqModel):
     channels:
     - config: DAPI
       exposure: 1.0
+    grid_plan:
+      columns: 2
+      rows: 2
     stage_positions:
     - x: 1.0
       y: 1.0
@@ -105,24 +110,34 @@ class MDASequence(UseqModel):
     metadata: Dict[str, Any] = Field(default_factory=dict)
     axis_order: str = "".join(INDICES)
     stage_positions: Tuple[Position, ...] = Field(default_factory=tuple)
+    grid_plan: AnyGridPlan = Field(default_factory=NoGrid)
     channels: Tuple[Channel, ...] = Field(default_factory=tuple)
     time_plan: AnyTimePlan = Field(default_factory=NoT)
     z_plan: AnyZPlan = Field(default_factory=NoZ)
 
     _uid: UUID = PrivateAttr(default_factory=uuid4)
     _length: Optional[int] = PrivateAttr(default=None)
+    _fov_size: Tuple[float, float] = PrivateAttr(default=(1, 1))
 
     @property
     def uid(self) -> UUID:
         """A unique identifier for this sequence."""
         return self._uid
 
+    def set_fov_size(self, fov_size: Tuple[float, float]) -> None:
+        """Set the field of view size.
+
+        This is used to calculate the number of positions in a grid plan.
+        """
+        self._fov_size = fov_size
+
     @no_type_check
     def replace(
         self,
         metadata: Dict[str, Any] = Undefined,
         axis_order: str = Undefined,
         stage_positions: Tuple[Position, ...] = Undefined,
+        grid_plan: AnyGridPlan = Undefined,
         channels: Tuple[Channel, ...] = Undefined,
         time_plan: AnyTimePlan = Undefined,
         z_plan: AnyZPlan = Undefined,
@@ -131,9 +146,11 @@ class MDASequence(UseqModel):
 
         MDASequences are immutable, so this method is useful for creating a new
         sequence with only a few fields changed.  The uid of the new sequence will
-        be different from the original
+        be different from the original.
         """
-        kwargs = {k: v for k, v in locals().items() if v is not Undefined}
+        kwargs = {
+            k: v for k, v in locals().items() if v is not Undefined and k != "self"
+        }
         state = self.dict(exclude={"uid"})
         return type(self)(**{**state, **kwargs})
 
@@ -180,8 +197,8 @@ class MDASequence(UseqModel):
                 z_plan=values.get("z_plan"),
                 stage_positions=values.get("stage_positions", ()),
                 channels=values.get("channels", ()),
+                grid_plan=values.get("grid_plan"),
             )
-
         return values
 
     def __eq__(self, other: Any) -> bool:
@@ -197,13 +214,14 @@ class MDASequence(UseqModel):
         z_plan: Optional[AnyZPlan] = None,
         stage_positions: Sequence[Position] = (),
         channels: Sequence[Channel] = (),
+        grid_plan: Optional[AnyGridPlan] = None,
     ) -> str:
         if (
             Z in order
             and POSITION in order
             and order.index(Z) < order.index(POSITION)
             and z_plan
-            and any(p.z_plan for p in stage_positions)
+            and any(p.sequence.z_plan for p in stage_positions if p.sequence)
         ):
             raise ValueError(
                 f"{Z!r} cannot precede {POSITION!r} in acquisition order if "
@@ -221,6 +239,28 @@ class MDASequence(UseqModel):
                 "{TIME!r} in the acquisition order: may not yield intended results."
             )
 
+        if (
+            GRID in order
+            and POSITION in order
+            and grid_plan
+            and not grid_plan.is_relative
+            and len(stage_positions) > 1
+        ):
+            sub_position_grid_plans = [
+                p for p in stage_positions if p.sequence and p.sequence.grid_plan
+            ]
+            if len(stage_positions) - len(sub_position_grid_plans) > 1:
+                warn("Global grid plan will override sub-position grid plans.")
+
+        if (
+            POSITION in order
+            and stage_positions
+            and any(p.sequence.stage_positions for p in stage_positions if p.sequence)
+        ):
+            raise ValueError(
+                "Currently, a Position sequence cannot have multiple stage positions!"
+            )
+
         return order
 
     def __str__(self) -> str:
@@ -240,7 +280,6 @@ class MDASequence(UseqModel):
         """Return the shape of this sequence.
 
         !!! note
-
             This doesn't account for jagged arrays, like skipped Z or channel frames.
         """
         return tuple(s for s in self.sizes.values() if s)
@@ -255,13 +294,18 @@ class MDASequence(UseqModel):
         """Single letter string of axes used in this sequence, e.g. `ztc`."""
         return "".join(k for k in self.axis_order if self.sizes[k])
 
-    def iter_axis(self, axis: str) -> Iterator[Position | Channel | float]:
+    def iter_axis(
+        self, axis: str
+    ) -> Iterator[Position | Channel | float | GridPosition]:
         """Iterate over the events of a given axis."""
         yield from {
             TIME: self.time_plan,
             POSITION: self.stage_positions,
             Z: self.z_plan,
             CHANNEL: self.channels,
+            GRID: self.grid_plan.iter_grid_positions(
+                self._fov_size[0], self._fov_size[1]
+            ),
         }[axis]
 
     def __iter__(self) -> Iterator[MDAEvent]:  # type: ignore [override]
@@ -311,13 +355,13 @@ class MDASequence(UseqModel):
 
 
 MDAEvent.update_forward_refs(MDASequence=MDASequence)
+Position.update_forward_refs(MDASequence=MDASequence)
 
 
 def iter_sequence(sequence: MDASequence) -> Iterator[MDAEvent]:
-    """Iterate over all events in the MDA sequence.
+    """Iterate over all events in the MDA sequence.'.
 
     !!! note
-
         This method will usually be used via [`useq.MDASequence.iter_events`][], or by
         simply iterating over the sequence.
 
@@ -334,23 +378,60 @@ def iter_sequence(sequence: MDASequence) -> Iterator[MDAEvent]:
     MDAEvent
         Each event in the MDA sequence.
     """
+    global_index = 0
     order = sequence.used_axes
 
     event_iterator = (enumerate(sequence.iter_axis(ax)) for ax in order)
-    for global_index, item in enumerate(product(*event_iterator)):
+    for item in product(*event_iterator):
         if not item:  # the case with no events
-            continue
+            continue  # pragma: no cover
 
         _ev = dict(zip(order, item))
         index = {k: _ev[k][0] for k in INDICES if k in _ev}
 
-        position: Optional[Position] = _ev[POSITION][1] if POSITION in _ev else None
-        channel: Optional[Channel] = _ev[CHANNEL][1] if CHANNEL in _ev else None
-        time: Optional[int] = _ev[TIME][1] if TIME in _ev else None
+        position = cast(
+            "Position | None", _ev[POSITION][1] if POSITION in _ev else None
+        )
+        channel = cast("Channel | None", _ev[CHANNEL][1] if CHANNEL in _ev else None)
+        time = cast("int | None", _ev[TIME][1] if TIME in _ev else None)
+        grid = cast("GridPosition | None", _ev[GRID][1] if GRID in _ev else None)
 
         # skip channels
         if channel and TIME in index and index[TIME] % channel.acquire_every:
             continue
+        # skip if also in position.sequence
+        if position and position.sequence:
+            if (
+                # if a position specifies channels, then the *global* channel index
+                # is no longer relevant... so we skip all but the first "global" channel
+                CHANNEL in index
+                and index[CHANNEL] != 0
+                # UNLESS the position specifies any other plan.
+                # NOTE: if we ever add more plans, they will need to be explicitly added
+                # https://github.com/pymmcore-plus/useq-schema/pull/85
+                and not any(
+                    (
+                        position.sequence.grid_plan,
+                        position.sequence.z_plan,
+                        position.sequence.time_plan,
+                    )
+                )
+            ):
+                continue
+            if Z in index and index[Z] != 0 and position.sequence.z_plan:
+                continue
+            if GRID in index and index[GRID] != 0 and position.sequence.grid_plan:
+                continue
+
+        _channel = (
+            _mda_event.Channel(config=channel.config, group=channel.group)
+            if channel
+            else None
+        )
+
+        _exposure = getattr(channel, "exposure", None)
+
+        pos_name = getattr(position, "name", None)
 
         try:
             z_pos = (
@@ -363,18 +444,109 @@ def iter_sequence(sequence: MDASequence) -> Iterator[MDAEvent]:
         except sequence._SkipFrame:
             continue
 
-        _channel = (
-            {"config": channel.config, "group": channel.group} if channel else None
-        )
+        if grid:
+            x_pos: Optional[float] = grid.x
+            y_pos: Optional[float] = grid.y
+            if grid.is_relative:
+                px = getattr(position, "x", 0) or 0
+                py = getattr(position, "y", 0) or 0
+                x_pos = x_pos + px if x_pos is not None else None
+                y_pos = y_pos + py if y_pos is not None else None
+        else:
+            x_pos = getattr(position, "x", None)
+            y_pos = getattr(position, "y", None)
+
+        if position and position.sequence:
+            for sub_event in iter_sequence(position.sequence):
+                # we're going to create a modified sub-event, inheriting some of the
+                # values from the parent event, and shifting the position of the
+                # event to account for global position offsets (or override if the
+                # sub-event has an absolute XYZ position plan.)
+                update_kwargs = dict(
+                    global_index=global_index,
+                    index={**index, **sub_event.index},
+                    sequence=sequence,
+                    pos_name=position.name or pos_name,
+                    **_maybe_shifted_positions(
+                        sub_event=sub_event,
+                        position=position,
+                        z_pos=z_pos,
+                        x_pos=x_pos,
+                        y_pos=y_pos,
+                    ),
+                )
+
+                # time, exposure, and channel are inherited from the parent event
+                # if they are not present on the sub-event.
+                for val, name in [
+                    (time, "min_start_time"),
+                    (_exposure, "exposure"),
+                    (_channel, "channel"),
+                ]:
+                    subval = getattr(sub_event, name)
+                    update_kwargs[name] = subval if subval is not None else val
+
+                yield sub_event.copy(update=update_kwargs)
+                global_index += 1
+
+            continue
+
         yield MDAEvent(
             index=index,
             min_start_time=time,
-            pos_name=getattr(position, "name", None),
-            x_pos=getattr(position, "x", None),
-            y_pos=getattr(position, "y", None),
+            pos_name=pos_name,
+            x_pos=x_pos,
+            y_pos=y_pos,
             z_pos=z_pos,
-            exposure=getattr(channel, "exposure", None),
+            exposure=_exposure,
             channel=_channel,
             sequence=sequence,
             global_index=global_index,
         )
+        global_index += 1
+
+
+def _maybe_shifted_positions(
+    sub_event: MDAEvent,  # the event we just created in the position sequence
+    position: Position,  # the position we are iterating
+    z_pos: float | None,  # global z position
+    x_pos: float | None,  # global x position
+    y_pos: float | None,  # global y position
+) -> dict:
+    kwargs = {}
+    # this function should only be called inside the sub-iteration loop
+    # so we can assume that position.sequence is not None
+    pos_seq = cast("MDASequence", position.sequence)
+
+    # if the position sequence has no z_plan, then we can use the global z_pos
+    # elif the position has a relative z_plan, then we need to shift the sub_event
+    if not pos_seq.z_plan:
+        kwargs["z_pos"] = z_pos
+    elif pos_seq.z_plan.is_relative:
+        kwargs["z_pos"] = _shift_axis("z", position.z, sub_event)
+
+    # if the position sequence has no grid_plane, then we can use the global z_pos
+    # elif the position has a relative grid_plan, then we need to shift the sub_event
+    if not pos_seq.grid_plan:
+        kwargs["x_pos"] = x_pos
+        kwargs["y_pos"] = y_pos
+    elif pos_seq.grid_plan.is_relative:
+        kwargs["x_pos"] = _shift_axis("x", position.x, sub_event)
+        kwargs["y_pos"] = _shift_axis("y", position.y, sub_event)
+    return kwargs
+
+
+def _shift_axis(axis: str, new_val: float | None, event: MDAEvent) -> float | None:
+    """Return a new value for the axis, accounting for the sub_event's position.
+
+    If both new_val and the corresponding axis position for the sub_event are None,
+    return None, otherwise return the sum of the two.
+    """
+    sub_pos = getattr(event, f"{axis}_pos")
+    return (
+        None
+        if new_val is None and sub_pos is None
+        else (new_val or 0) + sub_pos
+        if sub_pos is not None
+        else new_val
+    )
